package com.zillion.util.common;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.apache.http.util.EntityUtils;
import org.elasticsearch.action.admin.indices.create.CreateIndexRequest;
import org.elasticsearch.action.admin.indices.create.CreateIndexResponse;
import org.elasticsearch.action.admin.indices.delete.DeleteIndexRequest;
import org.elasticsearch.action.bulk.BulkRequest;
import org.elasticsearch.action.delete.DeleteRequest;
import org.elasticsearch.action.index.IndexRequest;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.action.support.master.AcknowledgedResponse;
import org.elasticsearch.action.update.UpdateRequest;
import org.elasticsearch.client.Request;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.Response;
import org.elasticsearch.client.RestClient;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.common.xcontent.XContentType;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.SearchHits;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.zillion.util.log.LogUtil;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.serializer.SerializerFeature;
import com.zillion.util.table.HTableUtil;
import com.zillion.util.table.ZillionTableColumn;
import com.zillion.util.table.ZillionTableColumnEnum;
import com.zillion.util.table.ZillionTableSchema;

public class ElasticSearchUtil {
	public static final String type_name = "g";

	public static void drop(RestHighLevelClient client, String Namespace, String table_name) throws Exception {
		DeleteIndexRequest request = new DeleteIndexRequest(Namespace + "_" + table_name);
		AcknowledgedResponse response = client.indices().delete(request, RequestOptions.DEFAULT);
		LogUtil.warn("ElasticSearch drop:" + Namespace + "_" + table_name + "\t" + response.isAcknowledged());
	}

	private static JSONObject generateES(ZillionTableSchema schema) {
		JSONObject result = new JSONObject();
		JSONObject properties = new JSONObject();
		Iterator<String> keyIter = schema.ColumnMap.keySet().iterator();
		while (keyIter.hasNext()) {
			String column = keyIter.next();
			ZillionTableColumn columnDef = schema.ColumnMap.get(column);
			JSONObject def = new JSONObject();
			if (columnDef.Type == ZillionTableColumnEnum.STRING || columnDef.Type == ZillionTableColumnEnum.COMPRESSSTRING) {
				def.put("type", "keyword");
			} else if (columnDef.Type == ZillionTableColumnEnum.LONG) {
				def.put("type", "long");
			} else if (columnDef.Type == ZillionTableColumnEnum.DOUBLE) {
				def.put("type", "double");
			} else if (columnDef.Type == ZillionTableColumnEnum.BOOLEAN) {
				def.put("type", "boolean");
			}
			properties.put(column, def);
		}
		result.put("properties", properties);
		return result;
	}

	public static void create(ZillionTableSchema schema, RestHighLevelClient client, String Namespace, String table_name) throws Exception {
		CreateIndexRequest request = new CreateIndexRequest(Namespace + "_" + table_name);
		request.settings(Settings.builder().put("index.number_of_shards", 1).put("index.number_of_replicas", 1));
		JSONObject wrapperObjects = generateES(schema);
		request.mapping(type_name, JSON.toJSONString(wrapperObjects, SerializerFeature.WriteMapNullValue), XContentType.JSON);
		CreateIndexResponse response = client.indices().create(request, RequestOptions.DEFAULT);
		LogUtil.warn("ElasticSearch create:" + Namespace + "_" + table_name + "\t" + response.isAcknowledged());
	}

	public static void insert(RestHighLevelClient client, String Namespace, String table_name, JSONArray InsertObjects) throws Exception {
		boolean bulk_insert = true;
		BulkRequest bulkRequest = new BulkRequest();
		for (int i = 0; i < InsertObjects.size(); i++) {
			JSONObject InsertObject = (JSONObject) InsertObjects.get(i);
			IndexRequest request = new IndexRequest(Namespace + "_" + table_name, type_name).source(InsertObject);
			if (bulk_insert) {
				bulkRequest.add(request);
			} else {
				client.index(request, RequestOptions.DEFAULT);
			}
		}
		if (bulk_insert) {
			client.bulk(bulkRequest, RequestOptions.DEFAULT);
		}
	}

	public static void delete(ZillionTableSchema schema, RestHighLevelClient client, String Namespace, String table_name, JSONArray InsertObjects)
			throws Exception {
		List<String> ids = select_ids(schema, client, Namespace, table_name, InsertObjects);
		BulkRequest bulkRequest = new BulkRequest();
		for (String id : ids) {
			DeleteRequest deleteRequest = new DeleteRequest(Namespace + "_" + table_name, type_name, id);
			bulkRequest.add(deleteRequest);
		}
		client.bulk(bulkRequest, RequestOptions.DEFAULT);
	}

	public static void update(ZillionTableSchema schema, RestHighLevelClient client, String Namespace, String table_name, JSONArray InsertObjects,
			JSONObject Set) throws Exception {
		List<String> ids = select_ids(schema, client, Namespace, table_name, InsertObjects);
		BulkRequest bulkRequest = new BulkRequest();
		for (String id : ids) {
			UpdateRequest deleteRequest = new UpdateRequest(Namespace + "_" + table_name, type_name, id);
			deleteRequest.doc(JSON.toJSONString(Set, SerializerFeature.WriteMapNullValue), XContentType.JSON);
			bulkRequest.add(deleteRequest);
		}
		client.bulk(bulkRequest, RequestOptions.DEFAULT);
	}

	public static List<String> select_ids(ZillionTableSchema schema, RestHighLevelClient client, String Namespace, String table_name,
			JSONArray InsertObjects) throws Exception {
		List<String> result = new ArrayList<String>();
		BoolQueryBuilder bqb = QueryBuilders.boolQuery();
		for (int i = 0; i < InsertObjects.size(); i++) {
			JSONObject InsertObject = (JSONObject) InsertObjects.get(i);
			BoolQueryBuilder bqbItem = QueryBuilders.boolQuery();
			for (String column : schema.Key) {
				bqbItem.must(QueryBuilders.termQuery(column, InsertObject.get(column)));
			}
			bqb.should(bqbItem);
		}
		SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
		sourceBuilder.query(bqb);
		sourceBuilder.timeout(new TimeValue(60, TimeUnit.SECONDS));
		String[] includeFields = schema.Key;
		String[] excludeFields = new String[] { "_type" };
		sourceBuilder.fetchSource(includeFields, excludeFields);

		SearchRequest searchRequest = new SearchRequest(Namespace + "_" + table_name);
		searchRequest.types(type_name);
		searchRequest.source(sourceBuilder);
		SearchResponse searchResponse = client.search(searchRequest, RequestOptions.DEFAULT);
		SearchHits hits = searchResponse.getHits();
		SearchHit[] searchHits = hits.getHits();
		for (SearchHit hit : searchHits) {
			result.add(hit.getId());
		}
		return result;
	}

	public static JSONArray selectAll(ZillionTableSchema schema, RestHighLevelClient client, String Namespace, String table_name) throws Exception {
		SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder();
		searchSourceBuilder.query(QueryBuilders.matchAllQuery());

		SearchRequest searchRequest = new SearchRequest(Namespace + "_" + table_name);
		searchRequest.types(type_name);
		searchRequest.source(searchSourceBuilder);
		SearchResponse searchResponse = client.search(searchRequest, RequestOptions.DEFAULT);
		JSONArray result = convert(schema, searchResponse);
		return result;
	}

	private static JSONArray convert(ZillionTableSchema schema, SearchResponse searchResponse) throws Exception {
		JSONArray result = new JSONArray();
		SearchHits hits = searchResponse.getHits();
		SearchHit[] searchHits = hits.getHits();
		for (SearchHit hit : searchHits) {
			JSONObject item = new JSONObject();
			Map<String, Object> datas = hit.getSourceAsMap();
			Iterator<String> keyIter = datas.keySet().iterator();
			while (keyIter.hasNext()) {
				String key = keyIter.next();
				ZillionTableColumn columnDef = schema.ColumnMap.get(key);
				Object value = HTableUtil.ConvertValue(columnDef, datas.get(key));
				item.put(key, value);
			}
			result.add(item);
		}
		return result;
	}

	public static JSONArray listIndex(RestHighLevelClient client) throws Exception {
		RestClient restClient = client.getLowLevelClient();
		try {
			// Response response = restClient.performRequest("GET", "/_cat/indices?v&format=json");
			Response response = restClient.performRequest(new Request("GET", "/_cat/indices?v&format=json"));
			String rawBody = EntityUtils.toString(response.getEntity());
			JSONArray Content = JSON.parseArray(rawBody);
			return Content;
		} catch (IOException e) {
			throw e;
		}
	}

	public static JSONObject clearIndex(RestHighLevelClient client, String index_name) throws Exception {
		RestClient restClient = client.getLowLevelClient();
		try {
			// Response response = restClient.performRequest("GET", "/_cat/indices?v&format=json");
			Request Request = new Request("POST", "/" + index_name + "/_delete_by_query");
			Request.setJsonEntity("{\"query\":{\"match_all\":{}}}");
			Response response = restClient.performRequest(Request);
			String rawBody = EntityUtils.toString(response.getEntity());
			JSONObject Content = JSON.parseObject(rawBody);
			return Content;
		} catch (IOException e) {
			throw e;
		}

	}
}
